// SPDX-License-Identifier: MIT
/**
 * Set key data from Edgelock se05x.
 * Copyright (c) 2022 Atmark Techno,Inc.
 */
#include <errno.h>
#include <getopt.h>
#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <se05x/fsl_sss_ftr.h>
#include <se05x/nxLog_sss.h> // for calling nLog() from ex_sss_main_inc.h
#include <se05x/ex_sss_boot.h>
#include <se05x/fsl_sss_se05x_apis.h>
#include <se05x/nxEnsure.h>
#include <se05x/se05x_const.h>
#include <se05x/ex_sss_objid.h>

#include <openssl/bn.h>
#include <openssl/ec.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <openssl/x509.h>
#include <openssl/obj_mac.h>

#ifndef OPENSSL_VERSION_NUMBER
#error "OPENSSL_VERSION_NUMBER not defined"
#endif
#if OPENSSL_VERSION_NUMBER < 0x10100000L
#error "not support older version than 1.1.0"
#endif

/* used by ex_sss_main_inc.h */
static ex_sss_boot_ctx_t gex_sss_boot_ctx;
#define EX_SSS_BOOT_PCONTEXT (&gex_sss_boot_ctx)
#define EX_SSS_BOOT_DO_ERASE 0
#define EX_SSS_BOOT_EXPOSE_ARGC_ARGV 1

#include <ex_sss_main_inc.h>

#if !defined(VERSION_STRING)
#define VERSION_STRING "for testing"
#endif
#define TYPE_STR_MAX 8 // max size of type option

static struct option options[] = { { "help", no_argument, 0, 'h' },
				   { "version", no_argument, 0, 'V' },
				   { "verbose", no_argument, 0, 'v' },
				   { "force", no_argument, 0, 'f' },
				   { "type", required_argument, 0, 't' },
				   { NULL, no_argument, NULL, 0 } };
static int use_verbose = 0;
static int forcibly_write = 0;

static void usage(FILE *stream)
{
	fputs("\
Usage: se05x_setkey [options] keyid file port\n",
	      stream);

	fputs("\
Store secure object to Edgelock se05x.\n",
	      stream);

	fputs("\
  keyid		   32bit hex string.\n\
  file		   file path of input file.\n\
  port		   device file and address for T=1 over I2C.\n\
		   must be last argument due to spec of plug and trust mw.\n",
	      stream);

	fputs("\
  options:\n\
  -f --force       write focibly secure object.\n\
  -t --type        format type\n\
                   specify input format:\n\
                     crt  x509 certificate (pem format).\n\
                     priv rsa/ecc key pair (pem format).\n\
  -V --version	   display version number.\n\
  -v --verbose	   output a diagnostic.\n\
  -h --help	   display help message.\n",
	      stream);
}

static void print_version(void)
{
	printf("se05x_setkey %s\n", VERSION_STRING);
	printf("copyright (c) 2022 Atmark Techno,Inc.\n");
}

static void print_openssl_err(const char *str_func)
{
	fprintf(stderr, "failed in openssl lib (%s). \n\n", str_func);
	ERR_print_errors_fp(stderr);
}

struct format {
	const char *type_str;
	int (*parse_func)(const char *, uint8_t *, uint32_t *, uint32_t *,
			  uint32_t *, size_t *);
};

static int load_keypair(const char *path, uint8_t *data, uint32_t *datalen,
			uint32_t *cipher_type, uint32_t *key_part,
			size_t *bitlen);
static int load_cert(const char *path, uint8_t *data, uint32_t *datalen,
		     uint32_t *cipher_type, uint32_t *key_part, size_t *bitlen);

/** supported format */
static struct format type_table[] = {
	{ "crt", load_cert },
	{ "priv", load_keypair },
	{ NULL, NULL },
};

static struct format *find_supported_format(const char *type_str)
{
	struct format *tbl;

	for (tbl = type_table; tbl->type_str; tbl++) {
		if (!strncmp(type_str, tbl->type_str, TYPE_STR_MAX)) {
			return tbl;
		}
	}
	return NULL;
}

static int try_loading_file_with_each_format(const char *path, uint8_t *data,
					     uint32_t *datalen,
					     uint32_t *cipher_type,
					     uint32_t *key_part, size_t *bitlen)
{
	struct format *tbl;

	for (tbl = type_table; tbl->type_str; tbl++) {
		int status;
		status = tbl->parse_func(path, data, datalen, cipher_type,
					 key_part, bitlen);
		if (status == kStatus_SSS_Success) {
			return kStatus_SSS_Success;
		}
	}

	*datalen = 0;
	*bitlen = 0;
	*cipher_type = kSSS_CipherType_NONE;
	*key_part = kSSS_KeyPart_NONE;

	return kStatus_SSS_Fail;
}

static int check_ec_priv_key(EVP_PKEY *pkey, uint32_t *cipher_type,
			     size_t *bitlen)
{
	const EC_KEY *ec;
	const EC_GROUP *group;
	int curve;

	if (use_verbose) {
		printf(" file format is ec private key.\n");
	}

	ec = EVP_PKEY_get0_EC_KEY(pkey);
	if (ec == NULL) {
		print_openssl_err("EVP_PKEY_get0_EC_KEY()");
		return kStatus_SSS_Fail;
	}

	group = EC_KEY_get0_group(ec);
	if (group == NULL) {
		print_openssl_err("EVP_PKEY_get0_EC_KEY()");
		return kStatus_SSS_Fail;
	}
	*bitlen = (uint32_t)EC_GROUP_order_bits(group);

	curve = EC_GROUP_get_curve_name(group);
	switch (curve) {
	case NID_X9_62_prime192v1:
	case NID_secp224r1:
	case NID_X9_62_prime256v1:
	case NID_secp384r1:
	case NID_secp521r1:
		*cipher_type = kSSS_CipherType_EC_NIST_P;
		break;
	default:
		fprintf(stderr, "not supported curve type. nid %d\n", curve);
		return kStatus_SSS_Fail;
	}

	if (use_verbose) {
		printf(" curve type %d\n", curve);
	}

	return kStatus_SSS_Success;
}

static int check_rsa_priv_key(EVP_PKEY *pkey, uint32_t *cipher_type,
			      size_t *bitlen)
{
	const RSA *rsa;

	if (use_verbose) {
		printf(" file format is rsa private key.\n");
	}

	rsa = EVP_PKEY_get0_RSA(pkey);
	if (rsa == NULL) {
		print_openssl_err("EVP_PKEY_get0_RSA()");
		return kStatus_SSS_Fail;
	}

	*bitlen = (uint32_t)RSA_bits(rsa);
	switch (*bitlen) {
	case 1024:
	case 2048:
	case 3072:
	case 4096:
		*cipher_type = kSSS_CipherType_RSA_CRT;
		break;
	default:
		fprintf(stderr, "not supported bit length. %d\n", 
			(uint32_t)*bitlen);
		return kStatus_SSS_Fail;
	}

	return kStatus_SSS_Success;
}

static int load_keypair(const char *path, uint8_t *data, uint32_t *datalen,
			uint32_t *cipher_type, uint32_t *key_part,
			size_t *bitlen)
{
	int result;
	FILE *fp = NULL;
	EVP_PKEY *pkey;

	if (use_verbose) {
		printf("try loading file as key pair.\n");
	}
	fp = fopen(path, "r");
	if (fp == NULL) {
		fprintf(stderr, " failed to open file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	pkey = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
	result = fclose(fp);
	if (result) {
		fprintf(stderr, " failed to close file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	if (pkey == NULL) {
		if (use_verbose) {
			printf(" faild to load file as key pair.\n");
		}
		return kStatus_SSS_Fail;
	}

	/** check detail of private key */
	int base_id = EVP_PKEY_base_id(pkey);

	switch (base_id) {
	case EVP_PKEY_EC:
		result = check_ec_priv_key(pkey, cipher_type, bitlen);
		break;
	case EVP_PKEY_RSA:
		result = check_rsa_priv_key(pkey, cipher_type, bitlen);
		break;
	default:
		fprintf(stderr, "not supported private key. %d\n", base_id);
		return kStatus_SSS_Fail;
	}
	if (result != kStatus_SSS_Success) {
		fprintf(stderr, "invalid private key file.\n");
		return kStatus_SSS_Fail;
	}
	*key_part = kSSS_KeyPart_Pair;

	/** convert to der binary */
	*datalen = (uint32_t)i2d_PrivateKey(pkey, &data);
	if (!*datalen) {
		print_openssl_err("i2d_PrivateKey()");
		return kStatus_SSS_Fail;
	}

	if (use_verbose) {
		printf(" result of reading PEM.\n");
		printf("  base_id      %d\n", base_id);
		printf("  cipher_type  %s\n", sss_cipher_type_sz(*cipher_type));
		printf("  bit length   %d\n", (uint32_t)*bitlen);
		printf("  data length  %d\n", *datalen);
	}

	return kStatus_SSS_Success;
}

static int load_cert(const char *path, uint8_t *data, uint32_t *datalen,
		     uint32_t *cipher_type, uint32_t *key_part, size_t *bitlen)
{
	int result;
	FILE *fp = NULL;
	X509 *x509;

	if (use_verbose) {
		printf("try loading file as certificate.\n");
	}
	fp = fopen(path, "r");
	if (fp == NULL) {
		fprintf(stderr, " failed to open file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	x509 = PEM_read_X509(fp, NULL, NULL, NULL);
	result = fclose(fp);
	if (result) {
		fprintf(stderr, " failed to close file. %s\n", strerror(errno));
		return kStatus_SSS_Fail;
	}
	if (x509 == NULL) {
		if (use_verbose) {
			printf(" faild to load file as certificate.\n");
		}
		return kStatus_SSS_Fail;
	}
	if (use_verbose) {
		printf(" file format is x509 cert.\n");
	}
	*cipher_type = kSSS_CipherType_Binary;
	*key_part = kSSS_KeyPart_Default;
	*bitlen = 0;

	/** convert to der binary */
	*datalen = (uint32_t)i2d_X509(x509, &data);
	if (!*datalen) {
		print_openssl_err("i2d_PrivateKey()");
		return kStatus_SSS_Fail;
	}

	return kStatus_SSS_Success;
}

sss_status_t ex_sss_entry(ex_sss_boot_ctx_t *pCtx)
{
	sss_status_t status = kStatus_SSS_Fail;
	int argc = gex_sss_argc;
	char *const *argv = (void *)gex_sss_argv;
	int c;
	char *endptr;

	uint64_t val;
	uint32_t keyid;
	char path_input[PATH_MAX];
	struct format *input_format = NULL;

	sss_object_t obj;
	uint8_t data[OPENSSL_RSA_MAX_MODULUS_BITS];
	uint32_t datalen = (uint32_t)sizeof(data);
	uint32_t cipher_type;
	uint32_t key_part;
	size_t bitlen;

	while ((c = getopt_long(argc, argv, "hVvft:", options, NULL)) != -1) {
		switch (c) {
		case 0:
			break;
		case 't':
			if (optarg) {
				input_format = find_supported_format(optarg);
				if (input_format == NULL) {
					fprintf(stderr,
						"not suppported format.\n");
					usage(stderr);
					return kStatus_SSS_Fail;
				}
			} else {
				fprintf(stderr,
					"type parameter is necessary.\n");
				usage(stderr);
				return kStatus_SSS_Fail;
			}
			break;
		case 'f':
			forcibly_write = 1;
			break;
		case 'h':
			usage(stdout);
			return kStatus_SSS_Success;
		case 'v':
			use_verbose = 1;
			printf("set verbose mode.\n");
			break;
		case 'V':
			print_version();
			return kStatus_SSS_Success;
		default:
			fprintf(stderr, "invalid option.\n");
			usage(stderr);
			return kStatus_SSS_Fail;
		}
	}
	if (optind != (argc - 3)) {
		fprintf(stderr, "incorrect number of arguments.\n");
		usage(stderr);
		return kStatus_SSS_Fail;
	}

	val = strtoul(argv[optind], &endptr, 16);
	if (!endptr || endptr[0] != '\0') {
		fprintf(stderr, "bad argument keyid %s needs to "
			"be a non-null hex integer\n", argv[optind]);
		usage(stderr);
		return kStatus_SSS_Fail;
	}
	if (val > UINT_MAX || val == 0) {
		fprintf(stderr, 
			"keyid %s needs to be within 32bit and not zero.\n",
			argv[optind]);
		return kStatus_SSS_Fail;
	}
	keyid = (uint32_t)val;

	strncpy(path_input, argv[optind + 1], PATH_MAX - 1);

	if (use_verbose) {
		printf("arguments\n");
		printf(" keyid	      0x%X\n", keyid);
		printf(" file	      %s\n", path_input);
		printf(" port	      %s\n", argv[optind + 2]);
	}

	/** initialize digital envelope library */
	OpenSSL_add_all_algorithms();
	ERR_load_BIO_strings();
	ERR_load_crypto_strings();

	/** load file */
	if (input_format == NULL) {
		status = try_loading_file_with_each_format(path_input, data,
							   &datalen,
							   &cipher_type,
							   &key_part, &bitlen);
	} else {
		status = input_format->parse_func(path_input, data, &datalen,
						  &cipher_type, &key_part,
						  &bitlen);
	}
	if (status != kStatus_SSS_Success) {
		fprintf(stderr, "not supported format.\n");
		goto exit;
	}

	/** erase already existed key if add force option */
	status = sss_key_object_init(&obj, &pCtx->ks);
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);

	status = sss_key_object_get_handle(&obj, keyid);
	if (status == kStatus_SSS_Success) {
		fprintf(stderr, "secure object (0x%X) exists.\n", keyid);
	}
	if (status == kStatus_SSS_Success && forcibly_write) {
		if (use_verbose) {
			printf("erase secure object (0x%x).\n", keyid);
		}
		status = sss_key_store_erase_key(&pCtx->ks, &obj);
		if (status != kStatus_SSS_Success) {
			fprintf(stderr, "failed to erase key.\n");
		}
		ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);
	}
	status = sss_key_object_init(&obj, &pCtx->ks);
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);

	status = sss_key_object_allocate_handle(&obj, keyid, key_part,
						cipher_type, datalen,
						kKeyObject_Mode_Persistent);
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);

	/** store secure object */
	status = sss_key_store_set_key(&pCtx->ks, &obj, data, datalen, bitlen,
				       NULL, 0);
	ENSURE_OR_GO_EXIT(status == kStatus_SSS_Success);

	if (use_verbose) {
		printf("succeeded.\n");
	}

	EVP_cleanup();

	return kStatus_SSS_Success;
exit:
	EVP_cleanup();
	fprintf(stdout, "failure occurred in sss api (%s)\n",
		sss_status_sz(status));
	return status;
}
